@@ -6,8 +6,9 @@ import numpy as np
 from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard
 from tensorflow import keras
 import os
+import yaml
 from typing import List, Optional, Tuple, Dict, Any
-from . import data, models, utils, predict, io, evaluate, neptune, data_hash  #, timeseries
+from . import data, models, utils, predict, io, evaluate, tracking, data_hash, augmentation  #, timeseries
 
 try:  # disabling eager execution speeds up everything
     from tensorflow.python.framework.ops import disable_eager_execution
@@ -25,20 +26,20 @@ except Exception as e:
 
 
 def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
-          save_dir: str = './', save_prefix: Optional[str] = None,
+          save_dir: str = './', save_prefix: Optional[str] = None, save_name: Optional[str] = None,
           model_name: str = 'tcn', nb_filters: int = 16, kernel_size: int = 16,
           nb_conv: int = 3, use_separable: List[bool] = False, nb_hist: int = 1024,
           ignore_boundaries: bool = True, batch_norm: bool = True,
           nb_pre_conv: int = 0, pre_nb_dft: int = 64,
           pre_kernel_size: int = 3, pre_nb_filters: int = 16, pre_nb_conv: int = 2,
-          upsample: bool = True,
+          upsample: bool = True, dilations: Optional[List[int]] = None,
           nb_lstm_units: int = 0,
           verbose: int = 2, batch_size: int = 32,
           nb_epoch: int = 400,
           learning_rate: Optional[float] = None, reduce_lr: bool = False, reduce_lr_patience: int = 5,
           fraction_data: Optional[float] = None, seed: Optional[int] = None, batch_level_subsampling: bool = False,
+          augmentations: Optional[str] = None,
           tensorboard: bool = False,
-          neptune_api_token: Optional[str] = None, neptune_project: Optional[str] = None,
           wandb_api_token: Optional[str] = None, wandb_project: Optional[str] = None, wandb_entity: Optional[str] = None,
           log_messages: bool = False, nb_stacks: int = 2, with_y_hist: bool = True,
           balance: bool = False, version_data: bool = True,
@@ -62,6 +63,10 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
                            Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + TIMESTAMP
                            or with SAVE_DIR/TIMESTAMP if SAVE_PREFIX is empty.
                            Defaults to '' (empty).
+        save_name (Optional[str]): Append to prefix.
+                           Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + SAVE_NAME
+                           or with SAVE_DIR/SAVE_NAME if SAVE_PREFIX is empty.
+                           Defaults to TIMESTAMP.
         model_name (str): Network architecture to use.
                           Use `tcn` (TCN) or `tcn_stft` (TCN with STFT frontend).
                           See das.models for a description of all models.
@@ -89,7 +94,8 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
                            if model is `tcn_tcn`: adds a frontend of N TCN blocks to the TCN.
                            if model is `tcn_stft`: adds a trainable STFT frontend.
                            Defaults to 0 (no frontend, no downsampling).
-        pre_nb_dft (int): Number of DFT points (roughly corresponding to number of Fourier filters) in the STFT frontend.
+        pre_nb_dft (int): Duration of filters (in samples) for the STFT frontend.
+                          Number of filters is pre_nb_dft // 2 + 1.
                           Defaults to 64.
         pre_nb_filters (int): Number of filters per layer in the pre-processing TCN.
                               Defaults to 16.
@@ -98,6 +104,7 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         upsample (bool): whether or not to restore the model output to the input samplerate.
                          Should generally be True during training and evaluation but my speed up inference.
                          Defaults to True.
+        dilations (List[int]): List of dilation rate, defaults to [1, 2, 4, 8, 16] (5 layer with 2x dilation per TCN block)
         nb_lstm_units (int): If >0, adds LSTM with `nb_lstm_units` LSTM units to the output of the stack of TCN blocks.
                              Defaults to 0 (no LSTM layer).
         verbose (int): Verbosity of training output (0 - no output during training, 1 - progress bar, 2 - one line per epoch).
@@ -123,11 +130,8 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         batch_level_subsampling (bool): Select fraction of data for training from random subset of shuffled batches.
                                         If False, select a continuous chunk of the recording.
                                         Defaults to False.
+        augmentations (Optional[str]): yaml file with augmentations. Defaults to None (no augmentations).
         tensorboard (bool): Write tensorboard logs to save_dir. Defaults to False.
-        neptune_api_token (Optional[str]): API token for logging to neptune.ai.
-                                           Defaults to None (no logging to neptune.ai).
-        neptune_project (Optional[str]): Project to log to for neptune.ai.
-                                         Defaults to None (no logging to neptune.ai).
         wandb_api_token (Optional[str]): API token for logging to wandb.
                                            Defaults to None (no logging to wandb).
         wandb_project (Optional[str]): Project to log to for wandb.
@@ -153,6 +157,9 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     if log_messages:
         logging.basicConfig(level=logging.INFO)
 
+    if dilations is None:
+        dilations = [1, 2, 4, 8, 16]
+
     # FIXME THIS IS NOT GREAT:
     sample_weight_mode = None
     data_padding = 0
@@ -169,11 +176,15 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         stride = 1  # should take every sample, since sampling rates of both x and y are now the same
         y_offset = int(round(nb_hist / 2))
 
+    if stride <= 0:
+        raise ValueError('Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries')
+
     if not upsample:
         output_stride = int(2**nb_pre_conv)
     else:
         output_stride = 1  # since we upsample output to original sampling rate. w/o upsampling: `output_stride = int(2**nb_pre_conv)` since each pre-conv layer does 2x max pooling
 
+
     if save_prefix is None:
         save_prefix = ''
 
@@ -182,9 +193,6 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     params = locals()
     del params['_qt_progress']
 
-    if stride <=0:
-        raise ValueError('Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries')
-
     # remove learning rate param if not set so the value from the model def is used
     if params['learning_rate'] is None:
         del params['learning_rate']
@@ -233,9 +241,19 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     else:
         shuffle_subset = None
 
+    if augmentations:
+        logging.info(f'Initializing augmentations from {augmentations}.')
+        aug_params = yaml.safe_load(open(augmentations, 'r'))
+        params['augmentations'] = aug_params
+        augs = augmentation.Augmentations.from_dict(params['augmentations'])
+        logging.info(f'   Got {len(augs)} augmentations.')
+    else:
+        augs = None
+
     data_gen = data.AudioSequence(d['train']['x'], d['train']['y'],
                                   shuffle=True, shuffle_subset=shuffle_subset,
                                   first_sample=first_sample_train, last_sample=last_sample_train, nb_repeats=1,
+                                  batch_processor=augs,
                                   **params)
     val_gen = data.AudioSequence(d['val']['x'], d['val']['y'],
                                  shuffle=False, shuffle_subset=shuffle_subset,
@@ -269,12 +287,15 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
 
     logging.info(model.summary())
     os.makedirs(os.path.abspath(save_dir), exist_ok=True)
-    save_name = '{0}/{1}{2}'.format(save_dir, save_prefix, time.strftime('%Y%m%d_%H%M%S'))
+    if save_name is None:
+        save_name = time.strftime('%Y%m%d_%H%M%S')
+    save_name = '{0}/{1}{2}'.format(save_dir, save_prefix, save_name)
     logging.info(f'Will save to {save_name}.')
 
     utils.save_params(params, save_name)
     checkpoint_save_name = save_name + "_model.h5"  # this will overwrite intermediates from previous epochs
 
+    # SET UP CALLBACKS
     callbacks = [ModelCheckpoint(checkpoint_save_name, save_best_only=True, save_weights_only=False, monitor='val_loss', verbose=1),
                  EarlyStopping(monitor='val_loss', patience=20),]
 
@@ -287,27 +308,11 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     if tensorboard:
         callbacks.append(TensorBoard(log_dir=save_dir))
 
-    del params['neptune_api_token']
-    if neptune_api_token and neptune_project:  # could also get those from env vars!
-        if not neptune.HAS_NEPTUNE:
-            logging.error('Could not import neptune in das.neptune.')
-        else:
-            try:
-                poseidon = neptune.Poseidon(neptune_project, neptune_api_token, params)
-                callbacks.append(poseidon.callback())
-            except Exception as e:
-                logging.exception('Neptune stuff failed.')
-
-    del params['wandb_api_token']
     if wandb_api_token and wandb_project:  # could also get those from env vars!
-        if not neptune.HAS_WANDB:
-            logging.error('Could not import neptune in das.neptune.')
-        else:
-            try:
-                magic = neptune.Magic(wandb_project, wandb_api_token, wandb_entity, params)
-                callbacks.append(magic.callback())
-            except Exception as e:
-                logging.exception('Neptune stuff failed.')
+        del params['wandb_api_token']
+        wandb = tracking.Wandb(wandb_project, wandb_api_token, wandb_entity, params)
+        if wandb:
+            callbacks.append(wandb.callback())
 
     # TRAIN NETWORK
     logging.info('start training')
@@ -340,8 +345,9 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params['class_names'], report_as_dict=True)
         logging.info(conf_mat)
         logging.info(report)
-        if neptune_api_token and neptune_project:  # could also get those from env vars!
-            poseidon.log_test_results(report)
+
+        if wandb_api_token and wandb_project:  # could also get those from env vars!
+            wandb.log_test_results(report)
 
         save_filename = "{0}_results.h5".format(save_name)
         logging.info('saving to ' + save_filename + '.')
