@@ -6,13 +6,20 @@ import time
 import logging
 import flammkuchen as fl
 import numpy as np
+
+try:
+    from tensorflow.python.keras.utils import tf_utils
+except ImportError:
+    from tensorflow.keras.utils import tf_utils
+
 from tensorflow.keras.callbacks import Callback, EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard
 from tensorflow import keras
 import keras_tuner as kt
 import yaml
 import os
+import sys
 from typing import List, Optional, Tuple, Dict, Any
-from . import data, models, utils, predict, io, evaluate, neptune, data_hash  #, timeseries
+from . import data, models, utils, predict, io, evaluate, tracking, data_hash  #, timeseries
 
 try:  # disabling eager execution speeds up everything
     from tensorflow.python.framework.ops import disable_eager_execution
@@ -34,22 +41,15 @@ class TunableModel(kt.HyperModel):
         self.params = params.copy()
         self.tune_config = tune_config
         if self.tune_config is None:
-            self.tune_config = {'nb_filters': [4, 8, 16, 32, 64, 128],
-                                'kernel_size': [4, 8, 16, 32, 64, 128],
+            self.tune_config = {'nb_filters': [4, 8, 16, 32, 64],
+                                'kernel_size': [4, 8, 16, 32, 64],
                                 'learning_rate': [0.01, 0.001, 0.0001],
-                                'nb_hist': [128, 256, 512, 1024, 2048, 4096, 8192],
-                                'nb_conv': [4, 8, 16, 32, 64, 128]}
+                                'nb_hist': [128, 256, 512, 1024, 2048],
+                                'nb_conv': [1, 2, 3, 4, 6, 8]}
 
     def build(self, hp):
-        if self.tune_config is not None:
-            for name, values in self.tune_config.items():
-                hp.Choice(name, values=values)
-        # else:  # defaults
-        #     hp.Choice('nb_filters', values=np.power(2, np.arange(3, 7)).tolist())
-        #     hp.Choice('kernel_size', values=np.power(2, np.arange(2, 7)).tolist())
-        #     hp.Choice('learning_rate', values=[1e-2, 1e-3, 1e-4])
-        #     hp.Choice('nb_hist', values=np.power(2, np.arange(7, 14)).tolist())
-        #     hp.Choice('nb_conv', values=np.arange(1, 7).tolist())
+        for name, values in self.tune_config.items():
+            hp.Choice(name, values=values)
 
         self.params.update(hp.values)
         model = models.model_dict['tcn'](**self.params)
@@ -65,10 +65,82 @@ class OracleCallback(Callback):
         self.tuner.on_epoch_end(self.tuner.current_trial, self.model, epoch, logs=logs)
 
 
+class ModelParamsCheckpoint(ModelCheckpoint):
+    """Callback to save the DAS model or model weights and parameters at some frequency."""
+
+    def _save_model(self, epoch=None, batch=None, logs=None):
+        """Saves the model.
+        Args:
+                epoch: the epoch this iteration is in.
+                batch: the batch this iteration is in. `None` if the `save_freq` is set to `epoch`.
+                logs: the `logs` dict passed in to `on_batch_end` or `on_epoch_end`.
+        """
+        logs = logs or {}
+
+        if isinstance(self.save_freq, int) or self.epochs_since_last_save >= self.period:
+            # Block only when saving interval is reached.
+            logs = tf_utils.sync_to_numpy_or_python_type(logs)
+            self.epochs_since_last_save = 0
+            filepath = self._get_file_path(epoch, batch, logs)
+
+            try:
+                if self.save_best_only:
+                    current = logs.get(self.monitor)
+                    if current is None:
+                        logging.warning('Can save best model only with %s available, '
+                                        'skipping.', self.monitor)
+                    else:
+                        if self.monitor_op(current, self.best):
+                            if self.verbose > 0:
+                                print('\nEpoch %05d: %s improved from %0.5f to %0.5f,'
+                                      ' saving model to %s' % (epoch + 1, self.monitor, self.best, current, filepath))
+                            self.best = current
+                            if self.save_weights_only:
+                                self.model.save_weights(filepath + '_model.h5', overwrite=True, options=self._options)
+                            else:
+                                self.model.save(filepath + '_model.h5', overwrite=True, options=self._options)
+                                # TODO: clean up params dict
+                                utils.save_params(self._normalize_tf_dict(self.model.params), filepath)
+                        else:
+                            if self.verbose > 0:
+                                print('\nEpoch %05d: %s did not improve from %0.5f' %
+                                      (epoch + 1, self.monitor, self.best))
+                else:
+                    if self.verbose > 0:
+                        print('\nEpoch %05d: saving model to %s' % (epoch + 1, filepath))
+                    if self.save_weights_only:
+                        self.model.save_weights(filepath + '_model.h5', overwrite=True, options=self._options)
+                    else:
+                        self.model.save(filepath + '_model.h5', overwrite=True, options=self._options)
+                        # TODO: clean up params dict
+                        utils.save_params(self._normalize_tf_dict(self.model.params), filepath)
+                self._maybe_remove_file()
+            except IsADirectoryError as e:  # h5py 3.x
+                raise IOError('Please specify a non-directory filepath for '
+                              'ModelCheckpoint. Filepath used is an existing '
+                              'directory: {}'.format(filepath))
+            except IOError as e:
+                # `e.errno` appears to be `None` so checking the content of `e.args[0]`.
+                if 'is a directory' in str(e.args[0]).lower():
+                    raise IOError('Please specify a non-directory filepath for '
+                                  'ModelCheckpoint. Filepath used is an existing '
+                                  'directory: {}'.format(filepath))
+                # Re-throw the error for any other causes.
+                raise e
+
+    def _normalize_tf_dict(self, d):
+        d = dict(d)
+        for key, val in d.items():
+            if isinstance(val, list):
+                d[key] = list(val)
+        return d
+
+
 class DasTuner(kt.Tuner):
 
-    def __init__(self, params, *args, **kwargs):
+    def __init__(self, params, *args, tracker=None, **kwargs):
         self.params = params.copy()
+        self.tracker = tracker
         super().__init__(*args, **kwargs)
 
     def run_trial(self, trial, train_x, train_y, val_x=None, val_y=None,
@@ -81,6 +153,9 @@ class DasTuner(kt.Tuner):
             callbacks.append(OracleCallback(self))
 
             self.params.update(trial.hyperparameters.values)
+
+            trial.metrics.update('val_loss', np.nan, step=0)
+            trial.status = 'INVALID'
             self.current_trial = trial
 
             # these need updating based on current hyperparameters
@@ -105,11 +180,21 @@ class DasTuner(kt.Tuner):
             if steps_per_epoch is None:
                 steps_per_epoch = min(len(data_gen), 1000)
 
+            if self.tracker is not None:
+                self.tracker.reinit(self.params)
             model = self.hypermodel.build(trial.hyperparameters)
+            model.params = self.params  # attach params to model so we can save them in ModelParamsCheckpoint
             model.fit(data_gen, validation_data=val_gen, epochs=epochs, steps_per_epoch=steps_per_epoch,
                       callbacks=callbacks, verbose=verbose, class_weight=class_weight)
+            if self.tracker is not None:
+                self.tracker.finish()
+            self.current_trial.status = 'RUNNING'
+        except KeyboardInterrupt:
+            logging.exception('Interrupted')
+            sys.exit(0)
         except:
             logging.exception("Something went wrong. Will try to continue.")
+            self.current_trial.status = 'INVALID'
 
 
 def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
@@ -119,15 +204,18 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
           ignore_boundaries: bool = True, batch_norm: bool = True,
           nb_pre_conv: int = 0, pre_nb_dft: int = 64,
           pre_kernel_size: int = 3, pre_nb_filters: int = 16, pre_nb_conv: int = 2,
+          upsample: bool = True, dilations: Optional[List[int]] = None,
           nb_lstm_units: int = 0,
           verbose: int = 2, batch_size: int = 32,
           nb_epoch: int = 400,
           learning_rate: Optional[float] = None, reduce_lr: bool = False, reduce_lr_patience: int = 5,
           fraction_data: Optional[float] = None, seed: Optional[int] = None, batch_level_subsampling: bool = False,
-          tensorboard: bool = False, neptune_api_token: Optional[str] = None, neptune_project: Optional[str] = None,
+          augmentations: str = None,
+          tensorboard: bool = False,
+          wandb_api_token: Optional[str] = None, wandb_project: Optional[str] = None, wandb_entity: Optional[str] = None,
           log_messages: bool = False, nb_stacks: int = 2, with_y_hist: bool = True,
           balance: bool = False, version_data: bool = True,
-          tune_config: Optional[str] = None,
+          tune_config: Optional[str] = None, nb_tune_trials: int = 1_000,
           _qt_progress: bool = False) -> Tuple[keras.Model, Dict[str, Any]]:
     """Tune the hyperparameters of a DAS network.
 
@@ -152,7 +240,6 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
                            Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + SAVE_NAME
                            or with SAVE_DIR/SAVE_NAME if SAVE_PREFIX is empty.
                            Defaults to TIMESTAMP.
-
         model_name (str): Network architecture to use.
                           Use `tcn` (TCN) or `tcn_stft` (TCN with STFT frontend).
                           See das.models for a description of all models.
@@ -180,12 +267,17 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
                            if model is `tcn_tcn`: adds a frontend of N TCN blocks to the TCN.
                            if model is `tcn_stft`: adds a trainable STFT frontend.
                            Defaults to 0 (no frontend, no downsampling).
-        pre_nb_dft (int): Number of DFT points (roughly corresponding to number of Fourier filters) in the STFT frontend.
+        pre_nb_dft (int): Duration of filters (in samples) for the STFT frontend.
+                          Number of filters is pre_nb_dft // 2 + 1.
                           Defaults to 64.
         pre_nb_filters (int): Number of filters per layer in the pre-processing TCN.
                               Defaults to 16.
         pre_kernel_size (int): Duration of filters (=kernels) in samples in the pre-processing TCN.
                                Defaults to 3.
+        upsample (bool): whether or not to restore the model output to the input samplerate.
+                         Should generally be True during training and evaluation but my speed up inference.
+                         Defaults to True.
+        dilations (List[int]): List of dilation rate, defaults to [1, 2, 4, 8, 16] (5 layer with 2x dilation per TCN block)
         nb_lstm_units (int): If >0, adds LSTM with `nb_lstm_units` LSTM units to the output of the stack of TCN blocks.
                              Defaults to 0 (no LSTM layer).
         verbose (int): Verbosity of training output (0 - no output during training, 1 - progress bar, 2 - one line per epoch).
@@ -212,20 +304,23 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
                                         If False, select a continuous chunk of the recording.
                                         Defaults to False.
         tensorboard (bool): Write tensorboard logs to save_dir. Defaults to False.
-        neptune_api_token (Optional[str]): API token for logging to neptune.ai.
-                                           Defaults to None (no logging to neptune.ai).
-        neptune_project (Optional[str]): Project to log to for neptune.ai.
-                                         Defaults to None (no logging to neptune.ai).
+        wandb_api_token (Optional[str]): API token for logging to wandb.
+                                           Defaults to None (no logging to wandb).
+        wandb_project (Optional[str]): Project to log to for wandb.
+                                         Defaults to None (no logging to wandb).
+        wandb_entity (Optional[str]): Entity to log to for wandb.
+                                        Defaults to None (no logging to wandb).
         log_messages (bool): Sets terminal logging level to INFO.
                              Defaults to False (will follow existing settings).
         nb_stacks (int): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to 2.
         with_y_hist (bool): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to True.
         balance (bool): Balance data. Weights class-wise errors by the inverse of the class frequencies.
-                        Defaults to False.G
+                        Defaults to False.
         version_data (bool): Save MD5 hash of the data_dir to log and params.yaml.
                              Defaults to True (set to False for large datasets since it can be slow).
         tune_config (Optional[str]): Yaml file with key:value pairs defining the search space for tuning.
                                      Keys are parameter names, values are lists of possible parameter values.
+        nb_tune_trials (int): Number of model variants to test during hyper parameter tuning. Defaults to 1_000.
 
         Returns
             model (keras.Model)
@@ -253,7 +348,13 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         stride = 1  # should take every sample, since sampling rates of both x and y are now the same
         y_offset = int(round(nb_hist / 2))
 
-    output_stride = 1  # since we upsample output to original sampling rate. w/o upsampling: `output_stride = int(2**nb_pre_conv)` since each pre-conv layer does 2x max pooling
+    if stride <= 0:
+        raise ValueError('Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries')
+
+    if not upsample:
+        output_stride = int(2**nb_pre_conv)
+    else:
+        output_stride = 1  # since we upsample output to original sampling rate. w/o upsampling: `output_stride = int(2**nb_pre_conv)` since each pre-conv layer does 2x max pooling
 
     if save_prefix is None:
         save_prefix = ''
@@ -336,24 +437,13 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     if save_name is None:
         save_name = time.strftime('%Y%m%d_%H%M%S')
     save_name = '{0}/{1}{2}'.format(save_dir, save_prefix, save_name)
+    params['save_name'] = save_name
+
     logging.info(f'Will save to {save_name}.')
 
-    tuner = DasTuner(
-        params=params,
-        oracle=kt.oracles.BayesianOptimization(
-            objective=kt.Objective("val_loss", "min"), max_trials=100,
-        ),
-        hypermodel=TunableModel(params, tune_config),
-        overwrite=True,
-        directory=save_dir,
-        project_name=os.path.basename(save_name),
-    )
-
-    logging.info(tuner.search_space_summary())
-
-    utils.save_params(params, save_name)
-
-    callbacks = [EarlyStopping(monitor='val_loss', patience=20),]
+    # setup callbacks
+    callbacks = [ModelParamsCheckpoint(save_name, save_best_only=True, save_weights_only=False, monitor='val_loss', verbose=1),
+                 EarlyStopping(monitor='val_loss', patience=15)]
 
     if reduce_lr:
         callbacks.append(ReduceLROnPlateau(patience=reduce_lr_patience, verbose=1))
@@ -364,20 +454,32 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
     if tensorboard:
         callbacks.append(TensorBoard(log_dir=save_name))
 
-    del params['neptune_api_token']
-    if neptune_api_token and neptune_project:  # could also get those from env vars!
-        if not neptune.HAS_NEPTUNE:
-            logging.error('Could not import neptune in das.neptune.')
-        else:
-            try:
-                poseidon = neptune.Poseidon(neptune_project, neptune_api_token, params)
-                callbacks.append(poseidon.callback())
-            except Exception as e:
-                logging.exception('Neptune stuff failed.')
+    if wandb_api_token and wandb_project:  # could also get those from env vars!
+            del params['wandb_api_token']
+            wandb = tracking.Wandb(wandb_project, wandb_api_token, wandb_entity, params)
+            if wandb:
+                callbacks.append(wandb.callback())
+    else:
+        wandb = None
+
+    tuner = DasTuner(
+        params=params,
+        oracle=kt.oracles.BayesianOptimization(
+            objective=kt.Objective("val_loss", "min"), max_trials=nb_tune_trials,
+        ),
+        hypermodel=TunableModel(params, tune_config),
+        overwrite=False,
+        directory=save_dir,
+        project_name=os.path.basename(save_name),
+        tracker=wandb,
+    )
+
+    logging.info(tuner.search_space_summary())
+    utils.save_params(params, save_name)
 
     # TRAIN NETWORK
     logging.info('Start hyperparameter tuning')
-    fit_hist = tuner.search(
+    tuner.search(
         train_x=d['train']['x'], train_y=d['train']['y'],
         val_x=d['val']['x'], val_y=d['val']['y'],
         epochs=nb_epoch,
@@ -385,7 +487,7 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         callbacks=callbacks,
         class_weight=params['class_weights'],
     )
-    tuner.results_summary()
+    logging.info(tuner.results_summary())
 
     # TEST
     if len(d['test']['x']) < nb_hist:
@@ -393,7 +495,7 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         return
     else:
         logging.info('re-loading last best model')
-        model = tuner.get_best_models()[0]
+        model, params = utils.load_model_and_params(params['save_name'])
 
         logging.info('predicting')
         # TODO: Need to update params with best hyperparams (e.g. nb-hist)
@@ -407,12 +509,12 @@ def train(*, data_dir: str, x_suffix: str = '', y_suffix: str = '',
         conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params['class_names'], report_as_dict=True)
         logging.info(conf_mat)
         logging.info(report)
-        if neptune_api_token and neptune_project:  # could also get those from env vars!
-            poseidon.log_test_results(report)
+        if wandb_api_token and wandb_project:  # could also get those from env vars!
+            wandb.log_test_results(report)
 
         save_filename = "{0}_results.h5".format(save_name)
         logging.info('saving to ' + save_filename + '.')
-        d = {'fit_hist': fit_hist.history,
+        d = {'fit_hist': [],
              'confusion_matrix': conf_mat,
              'classification_report': report,
              'x_test': x_test,
