@@ -10,12 +10,13 @@ def test_import():
     from das import io
     from das import make_dataset
     from das import models
-    from das import neptune
+    from das import tracking
     from das import npy_dir
     from das import predict
     from das import pulse_utils
     from das import segment_utils
     from das import train
+    from das import train_tune
     from das import utils
     from das import utils_plot
     from das import augmentation
